﻿using System.Windows;
using Carmenta.Engine;
using Point = Carmenta.Engine.Point;

namespace NauticalWaypointsAndEta
{
    /// <summary>
    /// Interaction logic for MapWindow
    /// </summary>
    public sealed partial class MapWindow : Window
    {
        private readonly StandardTool _standardTool = new StandardTool();
        private readonly StandardTouchTool _standardTouchTool = new StandardTouchTool();

        private CreateTool _createTool;
        private CreateTouchTool _createTouchTool;

        private Routing _routing;
        private MemoryDataSet _waypointsDataSet;
        private MemoryDataSet _itineraryDataSet;

        /// <summary>
        /// Initializes a new instance of the <see cref="MapWindow"/> class
        /// </summary>
        public MapWindow()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Loads the map configuration when the application starts
        /// </summary>
        /// <param name="sender">The object that raised the event</param>
        /// <param name="args">The event data</param>
        private void Window_Loaded(object sender, RoutedEventArgs args)
        {
            try
            {
                View view = LoadConfiguration("2d_view.px");
                if (view != null)
                {
                    // Set map control view and tools.
                    mapControl.View = view;
                    mapControl.Tool = _standardTool;
                    mapControl.TouchTool = _standardTouchTool;

                    _routing = new Routing();
                    _waypointsDataSet = view.FindChildObject("WaypointsMemoryDataSet") as MemoryDataSet;
                    _itineraryDataSet = view.FindChildObject("ItineraryMemoryDataSet") as MemoryDataSet;

                    // These tools will allow creating waypoint line features directly in the dataset.
                    _createTool = new CreateTool(_waypointsDataSet);
                    _createTouchTool = new CreateTouchTool(_waypointsDataSet);
                    _createTool.CreateMode = ToolCreateMode.Line;
                    _createTouchTool.CreateMode = ToolCreateMode.Line;
                    _createTool.FeatureCreated += OnWaypointsFeatureCreated;
                    _createTouchTool.FeatureCreated += OnWaypointsFeatureCreated;
                }
                else
                {
                    MessageBox.Show(
                        "No views found",
                        "NauticalWaypointsAndEta",
                        MessageBoxButton.OK,
                        MessageBoxImage.Error);
                    Close();
                }
            }
            catch (Exception ex)
            {
                // Display error message from exception.
                MessageBox.Show(
                    ex.Message,
                    "NauticalWaypointsAndEta",
                    MessageBoxButton.OK,
                    MessageBoxImage.Error);
                Close();
            }
        }

        /// <summary>
        /// Handles run-time errors in the map control
        /// </summary>
        /// <param name="sender">The object that raised the error event</param>
        /// <param name="args">The error data</param>
        private void MapControl_Error(object sender, Carmenta.Engine.Controls.ErrorEventArgs args)
        {
            // Disable view updates.
            if (mapControl.View != null)
            {
                mapControl.View.UpdateEnabled = false;
            }

            // Display error message.
            MessageBox.Show(
                args.Message,
                "NauticalWaypointsAndEta",
                MessageBoxButton.OK,
                MessageBoxImage.Error);

            // Mark the event as handled.
            args.Handled = true;
        }

        /// <summary>
        /// Loads the map configuration
        /// </summary>
        /// <param name="fileName">The map configuration file name</param>
        /// <returns>The first public view found in the map configuration</returns>
        private static View LoadConfiguration(string fileName)
        {
            // Load configuration file.
            Configuration configuration = new Configuration(fileName);

            // Get view from configuration.
            View view = null;
            foreach (string key in configuration.PublicNames)
            {
                view = configuration.GetPublicObject(key) as View;
                if (view != null)
                {
                    break;
                }
            }

            return view;
        }

        private void OnWaypointsFeatureCreated(object sender, FeatureCreatedEventArgs args)
        {
            // Reset to the navigation tool once the feature is created.
            mapControl.Tool = _standardTool;
            mapControl.TouchTool = _standardTouchTool;
        }

        private void CreateWaypoints_OnClick(object sender, RoutedEventArgs e)
        {
            // Clearing the dataset is not mandatory. You can work with several waypoints Features, for different vehicles, for example.
            // But in that case, make sure that the Routing class only uses 1 waypoints Feature in input
            // to the TerrainRouteOperator, in order to get the correct output Features for a specific vehicle.
            _waypointsDataSet.Clear();

            // Attach the create tools to the map control, to let the user:
            // - Tool: click on the map for creating waypoints and double-click to finish.
            // - TouchTool: draw the line on the map (this might create many waypoints, so you could create your own tool for this).
            mapControl.Tool = _createTool;
            mapControl.TouchTool = _createTouchTool;
        }

        private void CreateWaypointsCode_OnClick(object sender, RoutedEventArgs e)
        {
            // Clearing the dataset is not mandatory. You can work with several waypoints Features, for different vehicles, for example.
            // But in that case, make sure that the Routing class only uses 1 waypoints Feature in input
            // to the TerrainRouteOperator, in order to get the correct output Features for a specific vehicle.
            _waypointsDataSet.Clear();

            // Manually create the waypoints. Here we use hardcoded example positions,
            // but the waypoints could come from anywhere, like from the UI, from an application services, etc.
            PointCollection positions = new PointCollection
            {
                new Point(-82.786813, 27.671971),
                new Point(-82.620066, 27.646047),
                new Point(-82.56075, 27.775219)
            };
            Feature waypoints = new Feature(new LineGeometry(positions), _waypointsDataSet.Crs);

            // Replace waypoints dataset contents.
            using (new Guard(_waypointsDataSet))
            {
                _waypointsDataSet.Insert(waypoints);
            }

            // Update the view to show the new itinerary.
            mapControl.UpdateView();
        }

        private void CalculateItinerary_OnClick(object sender, RoutedEventArgs e)
        {
            // Calculate the itinerary based on the waypoints line.
            FeatureCollection itinerary = _routing.GetItinerary(out double eta, out string etaString);

            // Replace itinerary dataset contents.
            _itineraryDataSet.Clear();
            using (new Guard(_itineraryDataSet))
            {
                foreach (var itineraryLeg in itinerary)
                {
                    _itineraryDataSet.Insert(itineraryLeg);
                }
            }

            // Display ETA in UI using the already formatted 'etaString'.
            // You can format the 'eta' time in seconds into another string if you prefer.
            EtaString.Text = etaString;

            // Update the view to show the new itinerary.
            mapControl.UpdateView();
        }
    }
}
